#!/usr/bin/env python
import os,sys
import numpy as np
from netCDF4 import Dataset
from multiprocessing import Process

nday=float(sys.argv[1]) # tau_2 in Eq. 5: correlation length between uncertainties for different days
corh=int(sys.argv[2])   # tau_1 in Eq. 5: correlation length between undertainties for different hours
simuclmp=int(sys.argv[3]) # number of clumps per process
clmp0=int(sys.argv[4])  # the first clump in this simulation
hrand=float(sys.argv[5]) # random error for 3-h emissions, precomputed
annunc=0.3 # annual uncertainty

path='./'
resultpath=path+'AnnRes/'

year=2008
# AMS and ASS
if (nday==7 and corh==12 and hrand==0.49) or (nday==20 and corh==2400 and hrand==0.44):
  resultdir=resultpath+'CO2M_TmpInvBias_nday%1.1f_nhour%2.2i/'%(nday,corh)
else: # NoCor, MCS and SCS
  resultdir=resultpath+'CO2M_TmpInv_nday%1.1f_nhour%2.2i/'%(nday,corh)
if not os.path.isdir(resultdir):os.mkdir(resultdir)

clmp_cent=np.genfromtxt(path+'Clumps/Clumps_info.txt',delimiter=',')
noclmp=np.arange(len(clmp_cent))

def GlobeInv(startclmp):
  # Build temporal auto-correlation for hourly emissions within a day
  if corh>0:
    hcorrel_full=np.array([np.exp(-abs(ii-jj)*1./corh) for ii in range(24) for jj in range(24)]).reshape(24,24)
  elif corh==0:
    # No correlation for hourly emissions within a day
    hcorrel_full=np.diag(np.ones(24))
  # aggregate for 3-h and 21-h emissions
  v1=np.sum(hcorrel_full[8:11,8:11])
  v2=np.sum(hcorrel_full[8:11])-v1
  v3=np.sum(hcorrel_full)-v1-2*v2
  hcorrel_var=np.array([[v1/9,v2/3/21],[v2/3/21,v3/21/21]])
  Bhour=np.array(hcorrel_var)
  # Normalize
  Bhour=Bhour/Bhour[0,0]
  # Build temporal auto-correlation for emissions between different days
  if nday>0:
    Bday=np.array([np.exp(-abs(ii-jj)*1./nday) for ii in range(366) for jj in range(366)]).reshape(366,366)
  elif nday==0:
    # No correlation for emissions from different days
    Bday=np.diag(np.ones(366))
  # kronecker product of Bday and Bhour to consider all 3-h and 21-h emissions within a year, then multiplied by variance to get covariance matrix
  Bsubann=np.kron(Bday,Bhour)*hrand**2
  emis=np.array([3,21]*366).reshape(-1,1)
  # compute variance of annual component
  hbias2=((annunc*24*366)**2-np.sum(Bsubann*emis*emis.T))/24/366/24/366
  # full B matrix = sub-annual component + annual component 
  B=Bsubann+hbias2
  # do the inversion clump by clump
  for cc in range(clmp0+startclmp,np.min([clmp0+startclmp+simuclmp,len(noclmp)])):
    # if already computed, then skip this clump
    if os.path.isfile(resultdir+'A_Globe_clump%4.4i.npy'%clmp_cent[noclmp[cc],0]) and os.path.isfile(resultdir+'B_Globe_clump%4.4i.npy'%clmp_cent[noclmp[cc],0]):continue
    # Initialize pseudo-HRH
    HRH=np.zeros((366*2,366*2))
    for dd in range(1,367):
      for latitude in range(-50,75,10):
        for longitude in range(-175,180,10):
          if os.path.isfile(path+'Invmat/HRH_%3.3i_%3.3i_%3.3i.npz'%(dd,latitude,longitude)):
    	    clmpind=np.load(path+'Hmat/clmpindH_%3.3i_%3.3i_%3.3i.npy'%(dd,latitude,longitude))
            if noclmp[cc] in clmpind:
              # read HRH in the 10deg x 10deg on a given day
              HRH_win=np.load(path+'Invmat/HRH_%3.3i_%3.3i_%3.3i.npz'%(dd,latitude,longitude))['HRH']
              # find which index correspond to this clump
              index=np.argwhere(clmpind==noclmp[cc])
              # conduct the inversion for this day independently, accounting for spatial overlapping plumes from neighbouring clumps
              A_win=np.linalg.inv(np.linalg.inv(np.diag(np.ones(len(HRH_win)))*B[0,0])+HRH_win)
              # compute pseudo-HRH
              HRH_c=1./A_win[index,index]-1./B[0,0]
              # put pseudo-HRH in the correct place. If pseudo-HRH is larger than the value (default=zero) already there, it means the plumes are better sampled in this 10deg x 10deg window than previous one, then replace pseudo-HRH with the larger value
              HRH[(dd-1)*2,(dd-1)*2]=np.maximum(HRH[(dd-1)*2,(dd-1)*2],HRH_c)
    # annual inversion
    A=np.linalg.inv(np.linalg.inv(B)+HRH)
    # save result 
    A.dump(resultdir+'A_Globe_clump%4.4i.npy'%clmp_cent[noclmp[cc],0])
    B.dump(resultdir+'B_Globe_clump%4.4i.npy'%clmp_cent[noclmp[cc],0])
    (1.-np.sqrt(np.diag(A)/np.diag(B))).dump(resultdir+'UR_Globe_clump%4.4i.npy'%clmp_cent[noclmp[cc],0])
  return

if __name__=='__main__':
  process=[]
  for i in range(7):
    p=Process(target=GlobeInv,args=(i*simuclmp,))
    p.start()
    process.append(p)

  for p in process:
    p.join()

