#!/usr/bin/env python
import os,sys
import numpy as np
from netCDF4 import Dataset
from multiprocessing import Process

simuclmp=int(sys.argv[1]) # number of clumps computed in this simulation
clmp0=int(sys.argv[2]) # the first clump in this simulation
annunc=0.3 # annual uncertainty
# Industry
nday1=180.;corh1=2400
# Power
nday2=7.;corh2=24
# Transport
nday3=7.;corh3=12
# Other
nday4=14.;corh4=24

path='./'
resultpath=path+'AnnRes/'

year=2008
resultdir=resultpath+'CO2M_TmpInv_SectCS/'
if not os.path.isdir(resultdir):os.mkdir(resultdir)

clmp_cent=np.genfromtxt(path+'Clumps/Clumps_info.txt',delimiter=',')
noclmp=np.arange(len(clmp_cent))
# Build temporal auto-correlation for 3-h and 21-h emissions within a day for four sectors
hcorrel_full=np.array([np.exp(-abs(ii-jj)*1./corh1) for ii in range(24) for jj in range(24)]).reshape(24,24)
v1=np.sum(hcorrel_full[8:11,8:11])
v2=np.sum(hcorrel_full[8:11])-v1
v3=np.sum(hcorrel_full)-v1-2*v2
hcorrel_var=np.array([[v1/9,v2/3/21],[v2/3/21,v3/21/21]])
Bhour=np.array(hcorrel_var)
Bhour1=Bhour/Bhour[0,0]
hcorrel_full=np.array([np.exp(-abs(ii-jj)*1./corh2) for ii in range(24) for jj in range(24)]).reshape(24,24)
v1=np.sum(hcorrel_full[8:11,8:11])
v2=np.sum(hcorrel_full[8:11])-v1
v3=np.sum(hcorrel_full)-v1-2*v2
hcorrel_var=np.array([[v1/9,v2/3/21],[v2/3/21,v3/21/21]])
Bhour=np.array(hcorrel_var)
Bhour2=Bhour/Bhour[0,0]
hcorrel_full=np.array([np.exp(-abs(ii-jj)*1./corh3) for ii in range(24) for jj in range(24)]).reshape(24,24)
v1=np.sum(hcorrel_full[8:11,8:11])
v2=np.sum(hcorrel_full[8:11])-v1
v3=np.sum(hcorrel_full)-v1-2*v2
hcorrel_var=np.array([[v1/9,v2/3/21],[v2/3/21,v3/21/21]])
Bhour=np.array(hcorrel_var)
Bhour3=Bhour/Bhour[0,0]
hcorrel_full=np.array([np.exp(-abs(ii-jj)*1./corh4) for ii in range(24) for jj in range(24)]).reshape(24,24)
v1=np.sum(hcorrel_full[8:11,8:11])
v2=np.sum(hcorrel_full[8:11])-v1
v3=np.sum(hcorrel_full)-v1-2*v2
hcorrel_var=np.array([[v1/9,v2/3/21],[v2/3/21,v3/21/21]])
Bhour=np.array(hcorrel_var)
Bhour4=Bhour/Bhour[0,0]
# Build temporal auto-correlation for emissions between different days for four sectors
Bday1=np.array([np.exp(-abs(ii-jj)*1./nday1) for ii in range(366) for jj in range(366)]).reshape(366,366)
Bday2=np.array([np.exp(-abs(ii-jj)*1./nday2) for ii in range(366) for jj in range(366)]).reshape(366,366)
Bday3=np.array([np.exp(-abs(ii-jj)*1./nday3) for ii in range(366) for jj in range(366)]).reshape(366,366)
Bday4=np.array([np.exp(-abs(ii-jj)*1./nday4) for ii in range(366) for jj in range(366)]).reshape(366,366)
# kronecker product of Bday and Bhour to consider all 3-h and 21-h emissions within a year
B1=np.kron(Bday1,Bhour1)
B2=np.kron(Bday2,Bhour2)
B3=np.kron(Bday3,Bhour3)
B4=np.kron(Bday4,Bhour4)
emis=np.array([3,21]*366).reshape(-1,1)

def GlobeInv(startclmp):
  # the resolution of clump map is 1/120 deg
  clmpnc=Dataset(path+'Clumps/Clumps.nc','r')
  clmp=clmpnc.variables['clump']
  # the resolution of EDGAR emission map is 0.1 deg
  nc1=Dataset('inventory/EDGAR/edgar_v432_grids/EDGARv432_Industry.nc','r')
  emis1=nc1.variables['emission']
  nc2=Dataset('inventory/EDGAR/edgar_v432_grids/EDGARv432_Power.nc','r')
  emis2=nc2.variables['emission']
  nc3=Dataset('inventory/EDGAR/edgar_v432_grids/EDGARv432_Transport.nc','r')
  emis3=nc3.variables['emission']
  nc4=Dataset('inventory/EDGAR/edgar_v432_grids/EDGARv432_Other.nc','r')
  emis4=nc4.variables['emission']
  # loop for each clump
  for cc in range(clmp0+startclmp,np.min([clmp0+startclmp+simuclmp,len(noclmp)])):
    # if already computed, then skip this clump
    if os.path.isfile(resultdir+'A_Globe_clump%4.4i.npy'%clmp_cent[noclmp[cc],0]) and os.path.isfile(resultdir+'B_Globe_clump%4.4i.npy'%clmp_cent[noclmp[cc],0]):continue
    # locate the clump in the map
    loc=np.where(clmp[:]==clmp_cent[noclmp[cc],0])
    latbound1=np.min(loc[0])
    latbound2=np.max(loc[0])
    lonbound1=np.min(loc[1])
    lonbound2=np.max(loc[1])
    # cut the clump map and EDGAR emission map to small window to save memory cost; We assume emissions are uniformly distributed at all 1./120 grid boxes within an EDGAR 0.1 grid box
    clmpwin=clmp[latbound1:latbound2+1,lonbound1:lonbound2+1]
    emis1win=emis1[latbound1//12:latbound2//12+1,lonbound1//12:lonbound2//12+1]
    emis1win=np.repeat(emis1win,12,axis=0)
    emis1win=np.repeat(emis1win,12,axis=1)
    emis1win=emis1win[latbound1-latbound1//12*12:latbound2-latbound1//12*12+1,\
                      lonbound1-lonbound1//12*12:lonbound2-lonbound1//12*12+1]
    emis2win=emis2[latbound1//12:latbound2//12+1,lonbound1//12:lonbound2//12+1]
    emis2win=np.repeat(emis2win,12,axis=0)
    emis2win=np.repeat(emis2win,12,axis=1)
    emis2win=emis2win[latbound1-latbound1//12*12:latbound2-latbound1//12*12+1,\
                     lonbound1-lonbound1//12*12:lonbound2-lonbound1//12*12+1]
    emis3win=emis3[latbound1//12:latbound2//12+1,lonbound1//12:lonbound2//12+1]
    emis3win=np.repeat(emis3win,12,axis=0)
    emis3win=np.repeat(emis3win,12,axis=1)
    emis3win=emis3win[latbound1-latbound1//12*12:latbound2-latbound1//12*12+1,\
                      lonbound1-lonbound1//12*12:lonbound2-lonbound1//12*12+1]
    emis4win=emis4[latbound1//12:latbound2//12+1,lonbound1//12:lonbound2//12+1]
    emis4win=np.repeat(emis4win,12,axis=0)
    emis4win=np.repeat(emis4win,12,axis=1)
    emis4win=emis4win[latbound1-latbound1//12*12:latbound2-latbound1//12*12+1,\
                      lonbound1-lonbound1//12*12:lonbound2-lonbound1//12*12+1]
    mask=clmpwin==clmp_cent[noclmp[cc],0]
    tot=np.sum(emis1win[mask]+emis2win[mask]+emis3win[mask]+emis4win[mask])
    if tot!=0:
      frac1=np.sum(emis1win[mask])/tot
      frac2=np.sum(emis2win[mask])/tot
      frac3=np.sum(emis3win[mask])/tot
      frac4=np.sum(emis4win[mask])/tot
    else:
      # Clumps are defined based on ODIAC, but ODIAC does not separate four sectors. We use EDGAR to estimate the fractions of each sector. In case a clump in ODIAC does not have emissions in EDGAR map, we use the parameters for energy sector to compute its temporal auto-correlation
      frac1=0.;frac2=1.;frac3=0.;frac4=0.
    B=B1*frac1**2+B2*frac2**2+B3*frac3**2+B4*frac4**2
    # normalize B to achieve annual uncertainty = 0.3
    scale2=(annunc*24*366)**2/np.sum(B*emis*emis.T)
    B=B*scale2
    # Initialize pseudo-HRH
    HRH=np.zeros((366*2,366*2))
    for dd in range(1,367):
      for latitude in range(-50,75,10):
        for longitude in range(-175,180,10):
          if os.path.isfile(path+'Invmat/HRH_%3.3i_%3.3i_%3.3i.npz'%(dd,latitude,longitude)):
    	    clmpind=np.load(path+'Hmat/clmpindH_%3.3i_%3.3i_%3.3i.npy'%(dd,latitude,longitude))
            if noclmp[cc] in clmpind:
              # read HRH in the 10deg x 10deg on a given day
              HRH_win=np.load(path+'Invmat/HRH_%3.3i_%3.3i_%3.3i.npz'%(dd,latitude,longitude))['HRH']
              # find which index correspond to this clump
              index=np.argwhere(clmpind==noclmp[cc])
              # conduct the inversion for this day independently, accounting for spatial overlapping plumes from neighbouring clumps
              A_win=np.linalg.inv(np.linalg.inv(np.diag(np.ones(len(HRH_win)))*B[0,0])+HRH_win)
              # compute pseudo-HRH
              HRH_c=1./A_win[index,index]-1./B[0,0]
              # put pseudo-HRH in the correct place. If pseudo-HRH is larger than the value (default=zero) already there, it means the plumes are better sampled in this 10deg x 10deg window than previous one, then replace pseudo-HRH with the larger value
              HRH[(dd-1)*2,(dd-1)*2]=np.maximum(HRH[(dd-1)*2,(dd-1)*2],HRH_c)
    # annual inversion
    A=np.linalg.inv(np.linalg.inv(B)+HRH)
    # save result 
    A.dump(resultdir+'A_Globe_clump%4.4i.npy'%clmp_cent[noclmp[cc],0])
    B.dump(resultdir+'B_Globe_clump%4.4i.npy'%clmp_cent[noclmp[cc],0])
    (1.-np.sqrt(np.diag(A)/np.diag(B))).dump(resultdir+'UR_Globe_clump%4.4i.npy'%clmp_cent[noclmp[cc],0])
  return

if __name__=='__main__':
  process=[]
  for i in range(7):
    p=Process(target=GlobeInv,args=(i*simuclmp,))
    p.start()
    process.append(p)

  for p in process:
    p.join()

