#!/usr/bin/env python
import sys,os,calendar,datetime
import numpy as np
from scipy.sparse import csr_matrix
from netCDF4 import Dataset
from scipy.io import readsav
from multiprocessing import Process
# home-made functions
import earthcalc
import check

path='./'
satpath='./CO2M/'
respath='./'

year=2008

# overlapping boundaries of 10deg x 10deg windows = 500 km
tol=500000.
distlats=tol/earthcalc.latdist(0,1)

def Inversion(begday,endday):
  # read clump files (Wang et al., 2019, ESSD), spatial resolution is 1./120 deg
  clmp_cent=np.genfromtxt(path+'Clumps/Clumps_info.txt',delimiter=',')
  # clump index in column 0, lat of center in column 1, lon of center in column 2
  index=clmp_cent[:,0]
  clmpnc=Dataset(path+'Clumps/Clumps.nc','r')
  lats=clmpnc.variables['latitude'][:]
  lons=clmpnc.variables['longitude'][:]
  clmp=clmpnc.variables['clump'][:]
  # lats and lons -1./240 to move from the centers to the lower left edge
  llats=lats[:]+1./240  # 90.-1./240. to -90.+1./240.
  llons=lons[:]+1./240  # -180.+1./240. to 180.-1./240.
  # read hourly profile of scaling factors from TIMES, spatial resolution 0.25 deg
  hnc=Dataset(path+'inventory/ODIAC/hourly_scale_factors.nc','r')
  hlats=hnc.variables['latitude'][:]
  hlons=hnc.variables['longitude'][:]
  sf=hnc.variables['scaling_factors'] # 3D-array: 168*720*1440 
  ### extend hourly profile of Sunday, in case UTC<12 on Monday, LT corresponds to Sunday in Western Hemisphere
  sfpix=np.vstack((sf[-24:],sf[:]))
  # initialize current month
  curmon=0
  # open result file
  starttime=datetime.datetime.now()
  mdays=np.array(calendar.mdays)
  if calendar.isleap(year):mdays[2]+=1  # leap year
  # loop over the selected days (for parrallel computing)
  logfile=open(respath+'logfile','w',1)
  for dd in range(begday,endday):
    # initialization of the variables
    mon=np.searchsorted(np.cumsum(mdays),dd)
    mondd=dd-np.cumsum(mdays)[mon-1]
    satlats=np.array([]) # latitudes of satellite obs
    satlons=np.array([]) # longitudes of satellite obs
    re=np.array([])      # random errors of satellite obs, unit: ppb
    UR=np.zeros((len(clmp_cent)))  # result: uncertainty reduction
    bud=np.zeros((len(clmp_cent))) # budget of 3-hour emissions
    bud_day=np.zeros((len(clmp_cent))) # budget of daily emissions
    # read wind file
    if dd>begday:windnc.close()
    windnc=Dataset(path+'WindSpeed/Y%i/M%.2i/CCMP_Wind_Analysis_%i%.2i%.2i_V02.0_L3.0_RSS.nc'%(year,mon,year,mon,mondd),'r') # 6-hour wind fields, spatial resolution 0.25 deg
    wndlat=windnc.variables['latitude'][:]
    wndlon=windnc.variables['longitude'][:]
    uwnd=windnc.variables['uwnd'][:]
    vwnd=windnc.variables['vwnd'][:]
    # read emission file, if the new date is in next month
    if mon!=curmon:
      if curmon>0:emisnc.close()
      emisnc=Dataset(path+'inventory/ODIAC/odiac2016_1km_excl_intl_2015%.2i.nc'%mon,'r')
      emis=emisnc.variables['emission'][:]
      curmon=mon

    # Part 0 - Read coordinates and random errors of satellite obs for each day
    time0=datetime.datetime.now()
    direc=satpath+'%3.3i/'%dd
    check.display("direc="+direc,logfile)
    for f in sorted(os.listdir(direc)):
      clmpsat=[]
      data=readsav(direc+f)
      satlats=np.hstack((satlats,data['lat']))
      satlons=np.hstack((satlons,data['lon']))
      re=np.hstack((re,data['co2_re']))
      # window and UTC of the current sat
      minlat=np.min(data['lat']);maxlat=np.max(data['lat'])
      minlon=np.min(data['lon']);maxlon=np.max(data['lon'])
      obstime=f[22:26];hh=int(obstime[:2]);mm=int(obstime[2:])
      # find the clump centers located in a window close (<=500 km) to the track of the current to set hh of the clump
      maxabslat=np.max([abs(minlat),abs(maxlat)])
      distlons=tol/earthcalc.londist(maxabslat,0,1)
      clmpsat=np.argwhere((clmp_cent[:,1]>=(minlat-distlats))*(clmp_cent[:,1]<=(maxlat+distlats))*\
      	(clmp_cent[:,2]>=(minlon-distlons))*(clmp_cent[:,2]<=(maxlon+distlons)))
    check.display(' '.join(("satlats.shape,satlons.shape,re.shape:",str(satlats.shape),str(satlons.shape),str(re.shape))),logfile)
    # storage of the variables
    satlats.dump(respath+'sat/Satlats_%3.3i.npy'%dd)
    satlons.dump(respath+'sat/Satlons_%3.3i.npy'%dd)
    re.dump(respath+'sat/re_%3.3i.npy'%dd)
    check.display(' '.join(('Total time to read the sat obs:',str((datetime.datetime.now()-time0).seconds),'s')),logfile)

    # Part 1 - Computation of the response function for all clumps, one by one
    time1=datetime.datetime.now()
    check.display(' '.join(("Beginning of the computation of the response functions for all clumps:",str(time1))),logfile)
    for cc in range(len(index)):
      check.display(' '.join(('clump',str(cc),str(index[cc]))),logfile)
      # fine time zone of the clump and compute UTC time; Note: sfpix has been added 24 hours before Monday, here hh should +24
      timezone=np.searchsorted(np.arange(-172.5,180,15),clmp_cent[cc,2])-12
      timeclmp=datetime.datetime(year,mon,mondd,11,30)-datetime.timedelta(hours=timezone)  # UTC time
      hh=timeclmp.hour;mm=timeclmp.minute
      tind1=timeclmp.isoweekday()*24+hh
      tind2=tind1-3
      # coordinates of the pixels corresponding to the clump
      clmpmask=np.argwhere(clmp==index[cc])
      pixlats=llats[clmpmask[:,0]]
      pixlons=llons[clmpmask[:,1]]
      # coordinates of the pixels corresponding to the wind field
      pwndlat=np.searchsorted(wndlat[:]+0.125,pixlats)
      pwndlon=np.where(pixlons>0,np.searchsorted(wndlon[:]+0.125,pixlons),np.searchsorted(wndlon[:]+0.125,pixlons+360.))
      # computation of the wind direction and speed for each pixel
      wndvec=np.array([uwnd[hh//6,pwndlat,pwndlon],vwnd[hh//6,pwndlat,pwndlon]]).T # npix*2
      wnd=np.sqrt(np.sum(wndvec**2,axis=1))
      # coordinates of the scaling factors for each pixel
      hplat=np.searchsorted(hlats[:]+0.125,pixlats)
      hplon=np.searchsorted(hlons[:]+0.125,pixlons)
      # computation of the scaling factors for the three previous hours
      sfpixhr=sfpix[tind2:tind1+1,hplat,hplon]*np.array([[1.-mm/60.],[1.],[1.],[mm/60.]])
      aveemis=np.sum(sfpixhr,axis=0)/3.*emis[clmpmask[:,0],clmpmask[:,1]]/mdays[curmon]/24./3600.*1e6 # tonC/month -> gC/s
      # coefficients of the plume based on litterature (Ars et al., 2017, ACP)
      coef=np.where(wnd<4.,0.32,np.where((wnd>=4.)*(wnd<6.),0.22,0.16))
      # emission budget
      bud[np.where(index==index[cc])]=np.sum(aveemis)*3600*3./1e6  # ton
      bud_day[np.where(index==index[cc])]=np.sum(np.sum(sfpix[tind1-hh:tind1-hh+24,hplat,hplon],axis=0)\
        *emis[clmpmask[:,0],clmpmask[:,1]])/mdays[curmon]/24.  # ton
      # Computation of the response function for the selected clump
      if os.path.isfile(respath+'Hmat/Hclmp_'+str(dd)+'_'+str(int(index[cc]))+'.npz'):continue
      # Select sat obs located within 500 kms from the clump pixels
      minlat=np.min(pixlats);maxlat=np.max(pixlats);minlon=np.min(pixlons);maxlon=np.max(pixlons)
      maxabslat=np.max([abs(minlat),abs(maxlat)])
      distlons=tol/earthcalc.londist(maxabslat,0,1)
      satclmp=np.argwhere((satlats[:]>=(minlat-distlats))*(satlats[:]<=(maxlat+distlats))*\
        (satlons[:]>=(minlon-distlons))*(satlons[:]<=(maxlon+distlons)))
      if (len(satclmp))==0:continue
      H=np.zeros(len(satclmp)) # a clump value=0, means non-clump pixels
      # If nbpix*nbobs > 10000000, to avoid too much cost of memory, we compute response function for groups of 1000 source pixels
      if (len(clmpmask)*len(satclmp)>10000000):
        for ll in range(0,len(satclmp),1000):
	  ll2=np.min([ll+1000,len(satclmp)])
          # distance along meridians between source pixels and satellite obs
	  distlat=-earthcalc.latdist(satlats[satclmp[ll:ll2]].reshape(-1,1),pixlats)
          # distance along parallels between source pixels and satellite obs
          distlon=-earthcalc.londist(satlats[satclmp[ll:ll2]].reshape(-1,1),satlons[satclmp[ll:ll2]].reshape(-1,1),pixlons)
          # distance between source pixels and satellite obs
	  dists=np.sqrt(distlat**2+distlon**2)
          # directions between source pixels and satellite obs
	  vecs=np.dstack((distlon,distlat)) # shape: nobs*npix*2
          # distance along wind direction 
	  disti=np.minimum(dists,np.sum(vecs*wndvec,axis=2)/wnd)
          # distance perpendicular to wind direction
	  distj=np.where(dists>disti,np.sqrt(dists**2-disti**2),0)
          # if plume is not in the swath, skip 
	  if not ((disti>0.)*(distj<150000.)).any():continue
          # Gaussian plume model
	  sigy=np.where(disti>0,coef*disti/np.sqrt(1.+0.0004*disti),np.inf)
	  H[ll:ll2]=np.sum(np.where((disti<wnd*3*3600.)*(disti>0)*(distj<3*sigy),\
            aveemis/np.sqrt(2.*np.pi)/sigy/wnd*np.exp(-0.5*(distj/sigy)**2)/12./(1.013e5/9.81/29.e-3)*1e6,0),\
	    axis=1) # g/m2 -> ppm
      else:
        # distance along meridians between source pixels and satellite obs
	distlat=-earthcalc.latdist(satlats[satclmp[:]].reshape(-1,1),pixlats)
        # distance along parallels between source pixels and satellite obs
	distlon=-earthcalc.londist(satlats[satclmp[:]].reshape(-1,1),satlons[satclmp[:]].reshape(-1,1),pixlons)
        # distance between source pixels and satellite obs
	dists=np.sqrt(distlat**2+distlon**2)
        # directions between source pixels and satellite obs
	vecs=np.dstack((distlon,distlat))
        # distance along wind direction 
	disti=np.minimum(dists,np.sum(vecs*wndvec,axis=2)/wnd)
        # distance perpendicular to wind direction
	distj=np.where(dists>disti,np.sqrt(dists**2-disti**2),0)
        # if plume is not in the swath, skip 
	if not ((disti>0.)*(distj<150000.)).any():continue
        # Gaussian plume model
	sigy=np.where(disti>0,coef*disti/np.sqrt(1.+0.0004*disti),np.inf)
	H[:]=np.sum(np.where((disti<wnd*3*3600.)*(disti>0)*(distj<3*sigy),\
          aveemis/np.sqrt(2.*np.pi)/sigy/wnd*np.exp(-0.5*(distj/sigy)**2)/12./(1.013e5/9.81/29.e-3)*1e6,0),\
          axis=1) # g/m2 -> ppm
      
      # Store response function where at least one value > 0
      if len(np.argwhere(H[:]>0))==0:continue
      np.savez_compressed(respath+'Hmat/Hclmp_'+str(dd)+'_'+str(int(index[cc])),Hclmp=H)
      np.savez_compressed(respath+'Hmat/satclmp_'+str(dd)+'_'+str(int(index[cc])),satclmp=satclmp)
      np.savez_compressed(respath+'Hmat/reclmp_'+str(dd)+'_'+str(int(index[cc])),reclmp=re[satclmp[:]])
    check.display(' '.join(('Total time to compute the response functions for all clumps:',str((datetime.datetime.now()-time1).seconds))),logfile)
    bud.dump(respath+'bud/bud_'+str(dd)+'.npy')
    bud_day.dump(respath+'bud/bud_day_'+str(dd)+'.npy')

    # Part 2 - Computation of the H matrix for the inversion windows
    time2=datetime.datetime.now()
    check.display(' '.join(("Beginning of the construction of the H inversion matrix:",str(time2))),logfile)
    satlats=np.load(respath+'sat/Satlats_%3.3i.npy'%dd)
    satlons=np.load(respath+'sat/Satlons_%3.3i.npy'%dd)
    # latitude and longitude of center of 10deg x 10deg window
    for latitude in range(-50,75,10):
      for longitude in range(-175,180,10):
	# Clumps in each 10deg x 10deg window (each window having 500 km margins on four sides overlapping neighbouring windows)
        maxabslat=np.max([(latitude-5),abs(latitude+5)])
        distlons=tol/earthcalc.londist(maxabslat,0,1)
        clmpind=np.argwhere((clmp_cent[:,1]>=(latitude-5-distlats))*(clmp_cent[:,1]<=(latitude+5+distlats))*\
        		    (clmp_cent[:,2]>=(longitude-5-distlons))*(clmp_cent[:,2]<=(longitude+5+distlons)))
	# Clumps with response function
	clmpindH=np.array([],dtype='int')
	for cc in clmpind:
          if os.path.isfile(respath+'Hmat/Hclmp_'+str(dd)+'_'+str(int(index[cc]))+'.npz'):
            clmpindH=np.append(clmpindH,int(cc))
	# Selection of the satellite observations close to the considered clumps (distance within 500 km)
	if len(clmpindH)==0:continue  # if no clumps in the window
	minlat=np.min(clmp_cent[clmpindH[:],1]);maxlat=np.max(clmp_cent[clmpindH[:],1])
	minlon=np.min(clmp_cent[clmpindH[:],2]);maxlon=np.max(clmp_cent[clmpindH[:],2])
	maxabslat=np.max([abs(minlat),abs(maxlat)])
	distlons=tol/earthcalc.londist(maxabslat,0,1)
        satsind=np.argwhere((satlats[:]>minlat-distlats)*(satlats[:]<maxlat+distlats)*\
		            (satlons[:]>minlon-distlons)*(satlons[:]<maxlon+distlons))
	# Building of the positive response function for the considered clumps
	HclmpindH=np.zeros((len(satsind),len(clmpindH)))
	time3=datetime.datetime.now()
	for cc in clmpindH:
	  Hclmp=np.load(respath+'Hmat/Hclmp_'+str(dd)+'_'+str(int(index[cc]))+'.npz')['Hclmp']
          satclmp=np.load(respath+'Hmat/satclmp_'+str(dd)+'_'+str(int(index[cc]))+'.npz')['satclmp']
 	  reclmp=np.load(respath+'Hmat/reclmp_'+str(dd)+'_'+str(int(index[cc]))+'.npz')['reclmp']
          mask=np.argwhere((satlats[satclmp]>latitude-5-distlats)*(satlats[satclmp]<latitude+5+distlats)*\
                           (satlons[satclmp]>longitude-5-distlons)*(satlons[satclmp]<longitude+5+distlons))
          if np.prod(mask.shape)==0:continue # if clump is in the window, but its plume not, then skip
          Hclmp=Hclmp[mask[:,0]]
          satclmp=satclmp[mask[:,0]]
          reclmp=reclmp[mask[:,0]]
	  Hobsind=np.argwhere(Hclmp[:]>0)
          if np.prod(Hobsind.shape)==0:continue # if XCO2 enhacement not > 0, then remove
	  minsatlat=np.min(satlats[satclmp[Hobsind[:,0]]]);maxsatlat=np.max(satlats[satclmp[Hobsind[:,0]]])
	  minsatlon=np.min(satlons[satclmp[Hobsind[:,0]]]);maxsatlon=np.max(satlons[satclmp[Hobsind[:,0]]])
	  satsindclmp=np.argwhere((satlats[satsind[:]]>=minsatlat)*(satlats[satsind[:]]<=maxsatlat)*\
			          (satlons[satsind[:]]>=minsatlon)*(satlons[satsind[:]]<=maxsatlon))
	  indsatcc=np.argwhere(satsind[satsindclmp[:,0]]==satclmp[Hobsind[:,0],0])
	  if len(indsatcc)>0:
	    HclmpindH[satsindclmp[indsatcc[:,0],0],list(clmpindH).index(cc)]=Hclmp[Hobsind[indsatcc[:,1],0]]
        # remove zeros colums (plumes of clumps are not sampled by satellite obs) and rows (no clump generate plumes at the location of given satellite obs)
	Hposind1=np.argwhere(np.sum(HclmpindH,axis=1)>0)
	Hposind2=np.argwhere(np.sum(HclmpindH,axis=0)>0)
        if np.prod(Hposind1.shape)==0 or np.prod(Hposind2.shape)==0:continue # if all rows=0 and all columns=0, skip
        re=np.load(respath+'sat/re_%3.3i.npy'%dd)
	rem=re[satsind[Hposind1[:,0]]].flatten()
	# Saving data
        H=HclmpindH[Hposind1[:,0],:][:,Hposind2[:,0]]
        HR=csr_matrix(H.T/(rem/1000.)**2)
        HRH=HR.dot(H)
        np.savez_compressed(respath+'Hmat/H_%3.3i_%3.3i_%3.3i.npz'%(dd,latitude,longitude),H=H)
        rem.dump(respath+'Hmat/rem_%3.3i_%3.3i_%3.3i.npy'%(dd,latitude,longitude))
        clmpindH[Hposind2[:,0]].dump(respath+'Hmat/clmpindH_%3.3i_%3.3i_%3.3i.npy'%(dd,latitude,longitude))
        np.savez_compressed(respath+'Invmat/HRH_%3.3i_%3.3i_%3.3i.npz'%(dd,latitude,longitude),HRH=HRH)
    check.display(' '.join(("Total time for the construction of the H matrix:",str((datetime.datetime.now()-time2).seconds),'s')),logfile)
    
  clmpnc.close()
  hnc.close()
  return

if __name__=='__main__':
  process=[]
  nproc=int(sys.argv[1])
  startday=int(sys.argv[2])
  nday=int(sys.argv[3])
  print nproc,startday,nday
  for i in range(nproc):
    p=Process(target=Inversion,args=(startday+i*nday,np.min([startday+i*nday+nday,367])))
    p.start()
    process.append(p)

  for p in process:
    p.join()


